"""Main module."""
from functools import cached_property, lru_cache
from typing import Tuple, Union
import pm4py  # type: ignore
import static_frame as sf
from .two_set_helper import TwoSet
from .utils import (_validate_xes_dataframe_before_processing, _validate_activity_name,
                    _validate_case_id, _determine_start_end_per_case,
                    _remove_events_with_length_zero, _validate_xes_dataframe_after_processing)
from .damerau_levenshtein_helper import (_damerau_levenshtein_distancy_by_case,
                                         _levenshtein_distancy_by_case)
from .cross_correlation_helper import _cross_correlation_by_activity_case
from .two_set_helper import _two_set_by_activity_case
from .event_analysis_helper import (_event_analysis_by_activity_case, EventAnalysis)


class Context:
    """Class that offers main functionality of AquDeM.

    Both files are expected to be in the XES format, with special constraints:
    * The log must have an attribute specifying the sampling frequency in hertz
    (key: "sampling_freq") on the trace level (only the detected log).
    * Must use the concept:name,
    lifecycle:transition and time:timestamp standard extensions.
    * Each activity instance must have an event with at least
    the lifecycle transitions tart and complete.
    * In one case, the same activity can only be executed once at a time.

    An ACTIVITY_METRIC is a metric that is calculated for each activity type
    in each case separately.
    For requests that span multiple activities and/or cases, the results
    are aggregated.
    A SEQUENCE_METRIC is a metric that is calculated for each
    case separately.
    For requests that span multiple cases, the results are aggregated.

    :param str ground_truth: The ground truth log file path.
    :param str detected: The detected log file path.
    :return: An aqudem context instance,
    representing the comparison of two logs.
    """

    def __init__(self, ground_truth: str, detected: str):
        """Initialize the context with the ground truth and detected logs."""
        det_df = pm4py.read_xes(detected)
        gt_df = pm4py.read_xes(ground_truth)
        if det_df.empty or gt_df.empty:
            raise ValueError("One or more of logs is/are empty. Currently not supported.")
        base_gt = sf.FrameHE.from_pandas(gt_df.sort_values(by="time:timestamp"))
        base_det = sf.FrameHE.from_pandas(det_df.sort_values(by="time:timestamp"))
        _validate_xes_dataframe_before_processing(base_gt, base_det)
        self._ground_truth = _remove_events_with_length_zero(base_gt).relabel(sf.IndexAutoFactory)
        self._detected = _remove_events_with_length_zero(base_det).relabel(sf.IndexAutoFactory)
        _validate_xes_dataframe_after_processing(self._ground_truth, self._detected)
        self._sampling_freq: float = self._detected["case:sampling_freq"].iloc[0]
        self._start_end_per_case = _determine_start_end_per_case(self._ground_truth, self._detected)

    @property
    def ground_truth(self) -> sf.FrameHE:
        """Get the ground truth log."""
        return self._ground_truth

    @property
    def detected(self) -> sf.FrameHE:
        """Get the detected log."""
        return self._detected

    @property
    def sampling_freq(self) -> float:
        """Get the sampling frequency of the logs."""
        return self._sampling_freq

    @cached_property
    def activity_names(self) -> dict[str, list[str]]:
        """Extract all the available activity names from the XES logs.

        :return: A dictionary with "ground_truth" and "detected" keys, each
        containing a list of activity names.
        """
        return {
            "ground_truth": list(set(self._ground_truth["concept:name"].values)),
            "detected": list(set(self._detected["concept:name"].values))
        }

    @cached_property
    def case_ids(self) -> dict[str, list[str]]:
        """Extract all the available case IDs from the XES logs.

        :return: A dictionary with "ground_truth" and "detected" keys, each
        containing a list of case IDs.
        """
        return {
            "ground_truth": list(set(self._ground_truth["case:concept:name"].values)),
            "detected": list(set(self._detected["case:concept:name"].values))
        }

    @lru_cache(maxsize=20)
    def cross_correlation(self,
                          activity_name: str = "*",
                          case_id: str = "*") -> Tuple[float, float]:
        """Calculate the cross-correlation between the ground truth and detected logs.

        ACTIVITY_METRIC

        :param activity_name: The name of the activity to calculate the cross-correlation for.
            If "*" is passed, the cross-correlation will be calculated and averaged for all
            activities.
        :param case_id: The case ID to calculate the cross-correlation for.
            If "*" is passed, the cross-correlation will be calculated and averaged for all
            case IDs.
        :return: Tuple; first element: cross-correlation value, between 0 and 1.
            second element: relative shift to achieve maximum cross correlation.
        """
        _validate_activity_name(self._ground_truth,
                                self._detected,
                                activity_name)
        _validate_case_id(self._ground_truth,
                          self._detected,
                          case_id)
        return _cross_correlation_by_activity_case(self._ground_truth,
                                                   self._detected,
                                                   self._sampling_freq,
                                                   activity_name,
                                                   case_id,
                                                   self._start_end_per_case)

    @lru_cache(maxsize=20)
    def two_set(self, activity_name: str = "*", case_id: str = "*") -> TwoSet:
        """Calculate the 2SET metrics for a given activity. Absolute values.

        ACTIVITY_METRIC

        Includes the absolute and rate metrics, for details see the
        TwoSet class documentation.
        For more info on the metrics, refer to the metrics overview and/or:
        J. A. Ward, P. Lukowicz, and H. W. Gellersen, “Performance metrics for
        activity recognition,” ACM Trans. Intell. Syst. Technol., vol. 2, no. 1, pp. 1–23,
        Jan. 2011, doi: 10.1145/1889681.1889687.; 4.1.2

        :param activity_name: The name of the activity to calculate the two-set metrics for.
            If "*" is passed, the two-set metrics will be calculated
            and aggregated for all activities.
        :param case_id: The case ID to calculate the two-set metrics for.
            If "*" is passed, the two-set metrics will be calculated and
            aggregated for all case IDs.
        :return: A data class with the 2SET metrics.
        """
        _validate_activity_name(self._ground_truth,
                                self._detected,
                                activity_name)
        _validate_case_id(self._ground_truth,
                          self._detected,
                          case_id)
        return _two_set_by_activity_case(self._ground_truth,
                                         self._detected,
                                         self._sampling_freq,
                                         activity_name,
                                         case_id,
                                         self._start_end_per_case)


    @lru_cache(maxsize=20)
    def event_analysis(self, activity_name: str = "*", case_id: str = "*") -> EventAnalysis:
        """Calculate the EA metrics.

        ACTIVITY_METRIC

        Includes the absolute and rate metrics, for details see the
        EventAnalysis class documentation.
        For more info on the metrics, refer to the metrics overview and/or:
        J. A. Ward, P. Lukowicz, and H. W. Gellersen, “Performance metrics for
        activity recognition,” ACM Trans. Intell. Syst. Technol., vol. 2, no. 1, pp. 1–23,
        Jan. 2011, doi: 10.1145/1889681.1889687.; 4.2

        :param activity_name: The name of the activity to calculate the event analysis metrics for.
            If "*" is passed, the metrics will be calculated
            and aggregated for all activities.
        :param case_id: The case ID to calculate the event analysis metrics for.
            If "*" is passed, the metrics will be calculated and
            aggregated for all case IDs.
        :return: A data class with the EAD metrics.
        """
        _validate_activity_name(self._ground_truth,
                                self._detected,
                                activity_name)
        _validate_case_id(self._ground_truth,
                          self._detected,
                          case_id)
        return _event_analysis_by_activity_case(self._ground_truth,
                                                self._detected,
                                                activity_name,
                                                case_id,
                                                self._start_end_per_case)

    @lru_cache(maxsize=20)
    def damerau_levenshtein_distance(self, case_id: str = "*") -> Tuple[Union[float, int], float]:
        """Calculate the Damerau-Levenshtein distance between the ground truth and
            detected logs.

        SEQUENCE_METRIC

        Calculates both the absolute distance and the normalized distance.
        Order of activities based on start timestamps.

        :param case_id: The case ID to calculate the Damerau-Levenshtein distance for.
            If "*" is passed, the Damerau-Levenshtein distance will be calculated and
            averaged for all case IDs.
        :return: The Damerau-Levenshtein distance; tuple.
            The first value in the tuple represents the (average) absolute distance.
            The second value in the tuple represents the (average) normalized distance.
        """
        _validate_case_id(self._ground_truth, self._detected, case_id)
        return _damerau_levenshtein_distancy_by_case(
            self._ground_truth, self._detected, case_id)

    @lru_cache(maxsize=20)
    def levenshtein_distance(self, case_id: str = "*") -> Tuple[Union[float, int], float]:
        """Calculate the Levenshtein distance between the ground truth and detected logs.

        SEQUENCE_METRIC

        Calculates both the absolute distance and the normalized distance.
        Order of activities based on start timestamps.

        :param case_id: The case ID to calculate the Levenshtein distance for.
            If "*" is passed, the Levenshtein distance will be
            calculated and averaged for all case IDs.
        :return: The Levenshtein distance; tuple.
            The first value in the tuple represents the (average) absolute distance.
            The second value in the tuple represents the (average) normalized distance.
        """
        _validate_case_id(self._ground_truth, self._detected, case_id)
        return _levenshtein_distancy_by_case(
            self._ground_truth, self._detected, case_id)



"""Top-level package for AquDeM."""
from .aqudem import Context
from .event_analysis_helper import EventAnalysis
from .two_set_helper import TwoSet
from .utils import (XESMissingTraceNameAttribute, XESSamplingFreqError,
                    XESMissingTimestamp, XESMissingActivityName, XESMissingActivityInstance,
                    XESIncorrectLifecycleTransitionError)

__author__ = """ICS, Universität St. Gallen"""
__email__ = 'aaron.kurz@unisg.ch'
__version__ = '0.2.1'

__all__ = ["Context",
           "EventAnalysis",
           "TwoSet",
           "XESMissingTraceNameAttribute",
           "XESSamplingFreqError",
           "XESMissingTimestamp",
           "XESMissingActivityName",
           "XESMissingActivityInstance",
           "XESIncorrectLifecycleTransitionError"]



"""Tests for the two_set_helper module."""
import os
from dataclasses import fields
import aqudem
from aqudem.two_set_helper import (_two_set, TwoSet, _two_set_by_activity_case)
from .mocks.logs import (ground_truth_ten_eighteen, detected_ten_eighteen,
                         ground_truth_ten_eleven, detected_ten_eleven,
                         start_end_series_ten_eleven, start_end_series_ten_eighteen)

rate_props = ["uar", "oar", "tpr", "fr", "uor", "tnr", "oor", "ir", "dr", "mr"]


def verify_single_act_rates_ten_eighteen(result: TwoSet) -> None:
    assert result.uar == round(900 / 21601, 4)
    assert result.oar == round(900 / 7200, 4)
    assert result.tpr == round(13500 / 21601, 4)
    assert result.fr == round(3600 / 21601, 4)
    assert result.oor == round(3600 / 7200, 4)
    assert result.tnr == round(2700 / 7200, 4)
    assert result.uor == round(3601 / 21601, 4)
    assert result.ir == 0.0
    assert result.dr == 0.0
    assert result.mr == 0.0
    sum_metr = 0.0
    for field in rate_props:
        sum_metr += getattr(result, field)
    assert abs(sum_metr - 2.0) < 0.001


def verify_single_act_metrics_ten_eighteen(result: TwoSet) -> None:
    assert isinstance(result, TwoSet)
    assert result.ua == 900
    assert result.tp == 13500
    assert result.f == 3600
    assert result.oo == 3600
    assert result.oa == 900
    assert result.tn == 2700
    assert result.uo == 3601
    assert result.i == 0
    assert result.d == 0
    assert result.m == 0
    assert result.p == 21601
    assert result.n == 7200
    sum_metr = 0
    for field in fields(result):
        sum_metr += getattr(result, field.name)
    assert sum_metr == 28801


def verify_two_set_metrics_ten_eleven(result: TwoSet) -> None:
    assert isinstance(result, TwoSet)
    assert result.oa == 600
    assert result.tp == 1080
    assert result.f == 300
    assert result.uo == 300
    assert result.tn == 540
    assert result.oo == 421
    assert result.ua == 300
    assert result.i == 60
    assert result.d == 0
    assert result.m == 0
    assert result.t == 3601
    assert result.p == 1980
    assert result.n == 1621


def verify_two_set_rates_ten_eleven(result: TwoSet) -> None:
    assert isinstance(result, TwoSet)
    assert result.oar == round(600 / 1621, 4)
    assert result.tpr == round(1080 / 1980, 4)
    assert result.fr == round(300 / 1980, 4)
    assert result.uor == round(300 / 1980, 4)
    assert result.tnr == round(540 / 1621, 4)
    assert result.oor == round(421 / 1621, 4)
    assert result.uar == round(300 / 1980, 4)
    assert result.ir == round(60 / 1621, 4)
    assert result.dr == 0.0
    assert result.mr == 0.0


def verify_rates_avg_ten_eleven_eighteen(result: TwoSet) -> None:
    assert isinstance(result, TwoSet)
    assert result.uar == 0.0572
    assert result.oar == 0.0989
    assert result.tpr == 0.3843
    assert result.fr == 0.1041
    assert result.uor == 0.1041
    assert result.tnr == 0.1218
    assert result.oor == 0.121
    assert result.ir == 0.0083
    assert result.dr == 0.0
    assert result.mr == 0.0
    sum_metr = 0.0
    for field in rate_props:
        sum_metr += getattr(result, field)
    assert abs(sum_metr - 2.0) < 0.001


def verify_metrics_avg_ten_eleven_eighteen(result: TwoSet) -> None:
    assert isinstance(result, TwoSet)
    assert result.ua == 600
    assert result.tp == 7290
    assert result.f == 1950
    assert result.oo == 2010
    assert result.oa == 750
    assert result.tn == 1620
    assert result.uo == 1950
    assert result.i == 30
    assert result.d == 0
    assert result.m == 0


def test_two_set() -> None:
    result = _two_set(ground_truth_ten_eighteen,
                      detected_ten_eighteen,
                      start_end_series_ten_eighteen,
                      1)
    verify_single_act_metrics_ten_eighteen(result)

    result = _two_set(ground_truth_ten_eleven,
                      detected_ten_eleven,
                      start_end_series_ten_eleven,
                      1)
    verify_two_set_metrics_ten_eleven(result)


def test_two_set_rates() -> None:
    result = _two_set(ground_truth_ten_eighteen,
                      detected_ten_eighteen,
                      start_end_series_ten_eighteen,
                      1)
    verify_single_act_rates_ten_eighteen(result)

    result = _two_set(ground_truth_ten_eleven,
                      detected_ten_eleven,
                      start_end_series_ten_eleven,
                      1)
    verify_two_set_rates_ten_eleven(result)


def test_two_set_two_herz() -> None:
    result = _two_set(ground_truth_ten_eleven,
                      detected_ten_eleven,
                      start_end_series_ten_eleven,
                      2)
    assert isinstance(result, TwoSet)
    assert result.oa == 1200
    assert result.tp == 2160
    assert result.f == 600
    assert result.uo == 600
    assert result.tn == 1080
    assert result.oo == 841
    assert result.ua == 600
    assert result.i == 120
    assert result.d == 0
    assert result.m == 0
    assert result.t == 7201
    assert result.p == 3960
    assert result.n == 3241


def test_two_set_rates_two_hertz() -> None:
    result = _two_set(ground_truth_ten_eleven,
                      detected_ten_eleven,
                      start_end_series_ten_eleven,
                      2)
    assert isinstance(result, TwoSet)
    assert result.oar == round(1200 / 3241, 4)
    assert result.tpr == round(2160 / 3960, 4)
    assert result.fr == round(600 / 3960, 4)
    assert result.uor == round(600 / 3960, 4)
    assert result.tnr == round(1080 / 3241, 4)
    assert result.oor == round(841 / 3241, 4)
    assert result.uar == round(600 / 3960, 4)
    assert result.ir == round(120 / 3241, 4)
    assert result.dr == 0.0
    assert result.mr == 0.0


def test_two_set_by_activity_only_one_act_in_log() -> None:
    result_a = _two_set_by_activity_case(ground_truth_ten_eighteen,
                                         detected_ten_eighteen,
                                         1,
                                         "*",
                                         "1",
                                         start_end_series_ten_eighteen)
    verify_single_act_metrics_ten_eighteen(result_a)


def test_two_set_rates_by_activity_only_one_act_in_log() -> None:
    result_a = _two_set_by_activity_case(ground_truth_ten_eighteen,
                                         detected_ten_eighteen,
                                         1,
                                         "*",
                                         "1",
                                         start_end_series_ten_eighteen)
    verify_single_act_rates_ten_eighteen(result_a)


# pylint: disable=too-many-statements
def test_two_set_via_context_one_case_one_activity() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(activity_name="Activity A", case_id="ExampleTrace1")
    assert isinstance(res, TwoSet)
    assert res.tp == 780
    assert res.d == 0
    assert res.f == 60
    assert res.ua == 0
    assert res.uo == 120
    assert res.tn == 2281
    assert res.i == 0
    assert res.m == 0
    assert res.oa == 0
    assert res.oo == 300
    assert res.p == 960
    assert res.n == 2581

    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(activity_name="Activity B", case_id="ExampleTrace1")
    assert isinstance(res, TwoSet)
    assert res.tp == 390
    assert res.d == 0
    assert res.f == 0
    assert res.ua == 0
    assert res.uo == 0
    assert res.tn == 3001
    assert res.i == 0
    assert res.m == 30
    assert res.oa == 120
    assert res.oo == 0
    assert res.p == 390
    assert res.n == 3151

    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(activity_name="Activity C", case_id="ExampleTrace1")
    assert isinstance(res, TwoSet)
    assert res.tp == 421
    assert res.d == 240
    assert res.f == 0
    assert res.ua == 120
    assert res.uo == 0
    assert res.tn == 2760
    assert res.i == 0
    assert res.m == 0
    assert res.oa == 0
    assert res.oo == 0
    assert res.p == 781
    assert res.n == 2760

    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(activity_name="Activity A", case_id="ExampleTrace2")
    assert isinstance(res, TwoSet)
    assert res.tp == 0
    assert res.d == 120
    assert res.f == 0
    assert res.ua == 0
    assert res.uo == 0
    assert res.tn == 241
    assert res.i == 180
    assert res.m == 0
    assert res.oa == 0
    assert res.oo == 0
    assert res.p == 120
    assert res.n == 421

    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(activity_name="Activity B", case_id="ExampleTrace2")
    assert isinstance(res, TwoSet)
    assert res.tp == 300
    assert res.d == 0
    assert res.f == 60
    assert res.ua == 0
    assert res.uo == 61
    assert res.tn == 0
    assert res.i == 0
    assert res.m == 60
    assert res.oa == 60
    assert res.oo == 0
    assert res.p == 421
    assert res.n == 120


# pylint: disable=too-many-statements
def test_two_set_rates_via_context_one_case_one_activity() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(activity_name="Activity A", case_id="ExampleTrace1")
    assert isinstance(res, TwoSet)
    assert res.tpr == round(780 / 960, 4)
    assert res.dr == 0.0
    assert res.fr == round(60 / 960, 4)
    assert res.uar == 0.0
    assert res.uor == round(120 / 960, 4)
    assert res.tnr == round(2281 / 2581, 4)
    assert res.ir == 0.0
    assert res.mr == 0.0
    assert res.oar == 0.0
    assert res.oor == round(300 / 2581, 4)
    summed = 0.0
    for field in rate_props:
        summed += getattr(res, field)
    assert abs(summed - 2.0) < 0.001

    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(activity_name="Activity B", case_id="ExampleTrace1")
    assert isinstance(res, TwoSet)
    assert res.tpr == round(390 / 390, 4)
    assert res.dr == 0.0
    assert res.fr == 0.0
    assert res.uar == 0.0
    assert res.uor == 0.0
    assert res.tnr == round(3001 / 3151, 4)
    assert res.ir == 0.0
    assert res.mr == round(30 / 3151, 4)
    assert res.oar == round(120 / 3151, 4)
    assert res.oor == 0.0
    summed = 0.0
    for field in rate_props:
        summed += getattr(res, field)
    assert abs(summed - 2.0) < 0.001

    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(activity_name="Activity C", case_id="ExampleTrace1")
    assert isinstance(res, TwoSet)
    assert res.tpr == round(421 / 781, 4)
    assert res.dr == round(240 / 781, 4)
    assert res.fr == 0.0
    assert res.uar == round(120 / 781, 4)
    assert res.uor == 0.0
    assert res.tnr == round(2760 / 2760, 4)
    assert res.ir == 0.0
    assert res.mr == 0.0
    assert res.oar == 0.0
    assert res.oor == 0.0
    summed = 0.0
    for field in rate_props:
        summed += getattr(res, field)
    assert abs(summed - 2.0) < 0.001

    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(activity_name="Activity A", case_id="ExampleTrace2")
    assert isinstance(res, TwoSet)
    assert res.tpr == 0.0
    assert res.dr == round(120 / 120, 4)
    assert res.fr == 0.0
    assert res.uar == 0.0
    assert res.uor == 0.0
    assert res.tnr == round(241 / 421, 4)
    assert res.ir == round(180 / 421, 4)
    assert res.mr == 0.0
    assert res.oar == 0.0
    assert res.oor == 0.0
    summed = 0.0
    for field in rate_props:
        summed += getattr(res, field)
    assert abs(summed - 2.0) < 0.001

    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(activity_name="Activity B", case_id="ExampleTrace2")
    assert isinstance(res, TwoSet)
    assert res.tpr == round(300 / 421, 4)
    assert res.dr == 0.0
    assert res.fr == round(60 / 421, 4)
    assert res.uar == 0.0
    assert res.uor == round(61 / 421, 4)
    assert res.tnr == 0.0
    assert res.ir == 0.0
    assert res.mr == round(60 / 120, 4)
    assert res.oar == round(60 / 120, 4)
    assert res.oor == 0.0
    summed = 0.0
    for field in rate_props:
        summed += getattr(res, field)
    assert abs(summed - 2.0) < 0.001


def test_context_two_set_by_activty() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(activity_name="Activity A")
    assert isinstance(res, TwoSet)
    assert res.tp == round((780 + 0) / 2, 4)
    assert res.d == round((0 + 120) / 2, 4)
    assert res.f == round((60 + 0) / 2, 4)
    assert res.ua == round((0 + 0) / 2, 4)
    assert res.uo == round((120 + 0) / 2, 4)
    assert res.tn == round((2281 + 241) / 2, 4)
    assert res.i == round((0 + 180) / 2, 4)
    assert res.m == round((0 + 0) / 2, 4)
    assert res.oa == round((0 + 0) / 2, 4)
    assert res.oo == round((300 + 0) / 2, 4)
    assert res.p == round((960 + 120) / 2, 4)
    assert res.n == round((2581 + 421) / 2, 4)
    res = context.two_set(activity_name="Activity B")
    assert isinstance(res, TwoSet)
    assert res.tp == round((390 + 300) / 2, 4)
    assert res.d == round((0 + 0) / 2, 4)
    assert res.f == round((0 + 60) / 2, 4)
    assert res.ua == round((0 + 0) / 2, 4)
    assert res.uo == round((0 + 61) / 2, 4)
    assert res.tn == round((3001 + 0) / 2, 4)
    assert res.i == round((0 + 0) / 2, 4)
    assert res.m == round((30 + 60) / 2, 4)
    assert res.oa == round((120 + 60) / 2, 4)
    assert res.oo == round((0 + 0) / 2, 4)
    assert res.p == round((390 + 421) / 2, 4)
    assert res.n == round((3151 + 120) / 2, 4)
    res = context.two_set(activity_name="Activity C")
    assert isinstance(res, TwoSet)
    assert res.tp == 421
    assert res.d == 240
    assert res.f == 0
    assert res.ua == 120
    assert res.uo == 0
    assert res.tn == 2760
    assert res.i == 0
    assert res.m == 0
    assert res.oa == 0
    assert res.oo == 0
    assert res.p == 781
    assert res.n == 2760


def test_context_two_set_by_case() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(case_id="ExampleTrace1")
    assert isinstance(res, TwoSet)
    assert res.tp == round((780 + 390 + 421) / 3, 4)
    assert res.d == round((0 + 0 + 240) / 3, 4)
    assert res.f == round((60 + 0 + 0) / 3, 4)
    assert res.ua == round((0 + 0 + 120) / 3, 4)
    assert res.uo == round((120 + 0 + 0) / 3, 4)
    assert res.tn == round((2281 + 3001 + 2760) / 3, 4)
    assert res.i == round((0 + 0 + 0) / 3, 4)
    assert res.m == round((0 + 30 + 0) / 3, 4)
    assert res.oa == round((0 + 120 + 0) / 3, 4)
    assert res.oo == round((300 + 0 + 0) / 3, 4)
    assert res.p == round((960 + 390 + 781) / 3, 4)
    assert res.n == round((2581 + 3151 + 2760) / 3, 4)
    res = context.two_set(case_id="ExampleTrace2")
    assert isinstance(res, TwoSet)
    assert res.tp == round((0 + 300) / 2, 4)
    assert res.d == round((120 + 0) / 2, 4)
    assert res.f == round((0 + 60) / 2, 4)
    assert res.ua == round((0 + 0) / 2, 4)
    assert res.uo == round((0 + 61) / 2, 4)
    assert res.tn == round((241 + 0) / 2, 4)
    assert res.i == round((180 + 0) / 2, 4)
    assert res.m == round((0 + 60) / 2, 4)
    assert res.oa == round((0 + 60) / 2, 4)
    assert res.oo == round((0 + 0) / 2, 4)


def test_context_two_set_by_activty_case() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set()
    assert isinstance(res, TwoSet)
    assert res.tp == round((780 + 390 + 421 + 0 + 300) / 5, 4)
    assert res.d == round((0 + 0 + 240 + 120 + 0) / 5, 4)
    assert res.f == round((60 + 0 + 0 + 0 + 60) / 5, 4)
    assert res.ua == round((0 + 0 + 120 + 0 + 0) / 5, 4)
    assert res.uo == round((120 + 0 + 0 + 0 + 61) / 5, 4)
    assert res.tn == round((2281 + 3001 + 2760 + 241 + 0) / 5, 4)
    assert res.i == round((0 + 0 + 0 + 180 + 0) / 5, 4)
    assert res.m == round((0 + 30 + 0 + 0 + 60) / 5, 4)
    assert res.oa == round((0 + 120 + 0 + 0 + 60) / 5, 4)
    assert res.oo == round((300 + 0 + 0 + 0 + 0) / 5, 4)
    assert res.p == round((960 + 390 + 781 + 120 + 421) / 5, 4)
    assert res.n == round((2581 + 3151 + 2760 + 421 + 120) / 5, 4)


def test_context_two_set_rates_by_activty() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(activity_name="Activity A")
    assert isinstance(res, TwoSet)
    assert res.tpr == round((780 + 0) / (960 + 120), 4)
    assert res.dr == round((0 + 120) / (960 + 120), 4)
    assert res.fr == round((60 + 0) / (960 + 120), 4)
    assert res.uar == round((0 + 0) / (960 + 120), 4)
    assert res.uor == round((120 + 0) / (960 + 120), 4)
    assert res.tnr == round((2281 + 241) / (2581 + 421), 4)
    assert res.ir == round((0 + 180) / (2581 + 421), 4)
    assert res.mr == round((0 + 0) / (2581 + 421), 4)
    assert res.oar == round((0 + 0) / (2581 + 421), 4)
    assert res.oor == round((300 + 0) / (2581 + 421), 4)
    summed = 0.0
    for field in rate_props:
        summed += getattr(res, field)
    assert abs(summed - 2.0) < 0.001
    res = context.two_set(activity_name="Activity B")
    assert isinstance(res, TwoSet)
    assert res.tpr == round((390 + 300) / (390 + 421), 4)
    assert res.dr == round((0 + 0) / (390 + 421), 4)
    assert res.fr == round((0 + 60) / (390 + 421), 4)
    assert res.uar == round((0 + 0) / (390 + 421), 4)
    assert res.uor == round((0 + 61) / (390 + 421), 4)
    assert res.tnr == round((3001 + 0) / (3151 + 120), 4)
    assert res.ir == round((0 + 0) / (3151 + 120), 4)
    assert res.mr == round((30 + 60) / (3151 + 120), 4)
    assert res.oar == round((120 + 60) / (3151 + 120), 4)
    assert res.oor == round((0 + 0) / (3151 + 120), 4)
    summed = 0.0
    for field in rate_props:
        summed += getattr(res, field)
    assert abs(summed - 2.0) < 0.001
    res = context.two_set(activity_name="Activity C")
    assert isinstance(res, TwoSet)
    assert res.tpr == round(421 / 781, 4)
    assert res.dr == round(240 / 781, 4)
    assert res.fr == 0.0
    assert res.uar == round(120 / 781, 4)
    assert res.uor == 0.0
    assert res.tnr == round(2760 / 2760, 4)
    assert res.ir == 0.0
    assert res.mr == 0.0
    assert res.oar == 0.0
    assert res.oor == 0.0
    summed = 0.0
    for field in rate_props:
        summed += getattr(res, field)
    assert abs(summed - 2.0) < 0.001


def test_context_two_set_rates_by_case() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set(case_id="ExampleTrace1")
    assert isinstance(res, TwoSet)
    assert res.tpr == round((780 + 390 + 421) / (960 + 390 + 781), 4)
    assert res.dr == round((0 + 0 + 240) / (960 + 390 + 781), 4)
    assert res.fr == round((60 + 0 + 0) / (960 + 390 + 781), 4)
    assert res.uar == round((0 + 0 + 120) / (960 + 390 + 781), 4)
    assert res.uor == round((120 + 0 + 0) / (960 + 390 + 781), 4)
    assert res.tnr == round((2281 + 3001 + 2760) / (2581 + 3151 + 2760), 4)
    assert res.ir == round((0 + 0 + 0) / (2581 + 3151 + 2760), 4)
    assert res.mr == round((0 + 30 + 0) / (2581 + 3151 + 2760), 4)
    assert res.oar == round((0 + 120 + 0) / (2581 + 3151 + 2760), 4)
    assert res.oor == round((300 + 0 + 0) / (2581 + 3151 + 2760), 4)
    summed = 0.0
    for field in rate_props:
        summed += getattr(res, field)
    assert abs(summed - 2.0) < 0.001
    res = context.two_set(case_id="ExampleTrace2")
    assert isinstance(res, TwoSet)
    assert res.tpr == round((0 + 300) / (120 + 421), 4)
    assert res.dr == round((120 + 0) / (120 + 421), 4)
    assert res.fr == round((0 + 60) / (120 + 421), 4)
    assert res.uar == round((0 + 0) / (120 + 421), 4)
    assert res.uor == round((0 + 61) / (120 + 421), 4)
    assert res.tnr == round((241 + 0) / (421 + 120), 4)
    assert res.ir == round((180 + 0) / (421 + 120), 4)
    assert res.mr == round((0 + 60) / (421 + 120), 4)
    assert res.oar == round((0 + 60) / (421 + 120), 4)
    assert res.oor == round((0 + 0) / (421 + 120), 4)
    summed = 0.0
    for field in rate_props:
        summed += getattr(res, field)
    assert abs(summed - 2.0) < 0.001


def test_context_two_set_rates_by_activty_case() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set()
    assert isinstance(res, TwoSet)
    assert res.tpr == round((780 + 390 + 421 + 0 + 300) / (960 + 390 + 781 + 120 + 421), 4)
    assert res.dr == round((0 + 0 + 240 + 120 + 0) / (960 + 390 + 781 + 120 + 421), 4)
    assert res.fr == round((60 + 0 + 0 + 0 + 60) / (960 + 390 + 781 + 120 + 421), 4)
    assert res.uar == round((0 + 0 + 120 + 0 + 0) / (960 + 390 + 781 + 120 + 421), 4)
    assert res.uor == round((120 + 0 + 0 + 0 + 61) / (960 + 390 + 781 + 120 + 421), 4)
    assert res.tnr == round((2281 + 3001 + 2760 + 241 + 0) / (2581 + 3151 + 2760 + 421 + 120), 4)
    assert res.ir == round((0 + 0 + 0 + 180 + 0) / (2581 + 3151 + 2760 + 421 + 120), 4)
    assert res.mr == round((0 + 30 + 0 + 0 + 60) / (2581 + 3151 + 2760 + 421 + 120), 4)
    assert res.oar == round((0 + 120 + 0 + 60 + 0) / (2581 + 3151 + 2760 + 421 + 120), 4)
    assert res.oor == round((300 + 0 + 0 + 0 + 0) / (2581 + 3151 + 2760 + 421 + 120), 4)
    summed = 0.0
    for field in rate_props:
        summed += getattr(res, field)
    assert abs(summed - 2.0) < 0.001


def test_two_set_recall_precision_f1_balacc() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.two_set()
    assert isinstance(res, TwoSet)
    assert res.tpr == res.recall
    # merely check that the others do not raise an exception and that they are within
    # the range [0, 1]
    assert 0 <= res.recall <= 1
    assert 0 <= res.precision <= 1
    assert 0 <= res.f1 <= 1
    assert 0 <= res.balanced_accuracy <= 1



"""Tests for the event_analysis_helper module."""
import os
import aqudem
from aqudem.event_analysis_helper import (_event_analysis, EventAnalysis,
                                          _event_analysis_by_activity_case)
from .mocks.logs import (ground_truth_ten_eighteen, detected_ten_eighteen,
                         ground_truth_ten_eleven, detected_ten_eleven,
                         ground_truth_mixed_activity, detected_mixed_activity,
                         ground_truth_mixed_case, detected_mixed_case,
                         start_end_series_ten_eleven, start_end_series_ten_eighteen,
                         start_end_series_mixed_activity, start_end_series_mixed_case)


def _verify_ten_eleven_metrics(result: EventAnalysis) -> None:
    """ Verify the metrics for the 10-11 logs."""
    assert isinstance(result, EventAnalysis)
    assert result.d == 0
    assert result.f == 1
    assert result.fm == 0
    assert result.m == 0
    assert result.c == 4
    assert result.md == 0
    assert result.fmd == 0
    assert result.fd == 2
    assert result.id == 1
    assert result.total_gt_events == 3
    assert result.total_det_events == 5


def _verify_ten_eleven_rates(result: EventAnalysis) -> None:
    """ Verify the rates for the 10-11 logs."""
    assert isinstance(result, EventAnalysis)
    assert result.dr == 0.0
    assert result.fr == 0.3333
    assert result.fmr == 0.0
    assert result.mr == 0.0
    assert result.cr_gt == 0.6667
    assert result.mdr == 0.0
    assert result.fdr == 0.4
    assert result.idr == 0.2
    assert result.cr_det == 0.4


def _verify_ten_eighteen_metrics(result: EventAnalysis) -> None:
    """ Verify the metrics for the 10-18 logs."""
    assert isinstance(result, EventAnalysis)
    assert result.d == 0
    assert result.f == 2
    assert result.fm == 0
    assert result.m == 0
    assert result.c == 2
    assert result.md == 0
    assert result.fmd == 0
    assert result.fd == 4
    assert result.id == 0
    assert result.total_gt_events == 3
    assert result.total_det_events == 5


def _verify_ten_eighteen_rates(result: EventAnalysis) -> None:
    """ Verify the rates for the 10-18 logs."""
    assert isinstance(result, EventAnalysis)
    assert result.dr == 0.0
    assert result.fr == 0.6667
    assert result.fmr == 0.0
    assert result.mr == 0.0
    assert result.cr_gt == 0.3333
    assert result.mdr == 0.0
    assert result.fdr == 0.8
    assert result.idr == 0.0
    assert result.cr_det == 0.2


def test_event_analysis_ten_eleven() -> None:
    """ Test the event analysis for the 10-11 logs."""
    result = _event_analysis(ground_truth_ten_eleven,
                             detected_ten_eleven,
                             start_end_series_ten_eleven)
    _verify_ten_eleven_metrics(result)


def test_event_analysis_rates_ten_eleven() -> None:
    """ Test the event analysis rates for the 10-11 logs."""
    result = _event_analysis(ground_truth_ten_eleven,
                             detected_ten_eleven,
                             start_end_series_ten_eleven)
    _verify_ten_eleven_rates(result)


def test_event_analysis_ten_eighteen() -> None:
    result = _event_analysis(ground_truth_ten_eighteen,
                             detected_ten_eighteen,
                             start_end_series_ten_eighteen)
    _verify_ten_eighteen_metrics(result)


def test_event_analysis_rates_ten_eighteen() -> None:
    result = _event_analysis(ground_truth_ten_eighteen,
                             detected_ten_eighteen,
                             start_end_series_ten_eighteen)
    _verify_ten_eighteen_rates(result)


def test_event_analysis_by_activity() -> None:
    result = _event_analysis_by_activity_case(ground_truth_mixed_activity,
                                              detected_mixed_activity,
                                              "A",
                                              "1",
                                              start_end_series_mixed_activity)
    _verify_ten_eleven_metrics(result)
    result = _event_analysis_by_activity_case(ground_truth_mixed_activity,
                                              detected_mixed_activity,
                                              "B",
                                              "1",
                                              start_end_series_mixed_activity)
    _verify_ten_eighteen_metrics(result)
    result = _event_analysis_by_activity_case(ground_truth_mixed_activity,
                                              detected_mixed_activity,
                                              "*",
                                              "1",
                                              start_end_series_mixed_activity)
    assert isinstance(result, EventAnalysis)
    assert result.d == 0
    assert result.f == 1.5
    assert result.fm == 0
    assert result.m == 0
    assert result.c == 3
    assert result.md == 0
    assert result.fmd == 0
    assert result.fd == 3
    assert result.id == 0.5
    assert result.total_gt_events == 3
    assert result.total_det_events == 5


def test_event_analysis_rates_by_activity() -> None:
    result = _event_analysis_by_activity_case(ground_truth_mixed_activity,
                                              detected_mixed_activity,
                                              "A",
                                              "1",
                                              start_end_series_mixed_activity)
    _verify_ten_eleven_rates(result)
    result = _event_analysis_by_activity_case(ground_truth_mixed_activity,
                                              detected_mixed_activity,
                                              "B",
                                              "1",
                                              start_end_series_mixed_activity)
    _verify_ten_eighteen_rates(result)
    result = _event_analysis_by_activity_case(ground_truth_mixed_activity,
                                              detected_mixed_activity,
                                              "*",
                                              "1",
                                              start_end_series_mixed_activity)
    assert isinstance(result, EventAnalysis)
    assert result.dr == 0.0
    assert result.fr == 0.5
    assert result.fmr == 0.0
    assert result.mr == 0.0
    assert result.cr_gt == 0.5
    assert result.mdr == 0.0
    assert result.fdr == 0.6
    assert result.idr == 0.1
    assert result.cr_det == 0.3


def test_event_analysis_by_activity_case() -> None:
    result = _event_analysis_by_activity_case(ground_truth_mixed_case,
                                              detected_mixed_case,
                                              "*",
                                              "1",
                                              start_end_series_mixed_case)
    _verify_ten_eleven_metrics(result)
    result = _event_analysis_by_activity_case(ground_truth_mixed_case,
                                              detected_mixed_case,
                                              "*",
                                              "2",
                                              start_end_series_mixed_case)
    _verify_ten_eighteen_metrics(result)
    result = _event_analysis_by_activity_case(ground_truth_mixed_case,
                                              detected_mixed_case,
                                              "*",
                                              "*",
                                              start_end_series_mixed_case)
    assert isinstance(result, EventAnalysis)
    assert result.d == 0
    assert result.f == 1.5
    assert result.fm == 0
    assert result.m == 0
    assert result.c == 3
    assert result.md == 0
    assert result.fmd == 0
    assert result.fd == 3
    assert result.id == 0.5
    assert result.total_gt_events == 3
    assert result.total_det_events == 5


def test_event_analysis_rates_by_activity_case() -> None:
    result = _event_analysis_by_activity_case(ground_truth_mixed_case,
                                              detected_mixed_case,
                                              "*",
                                              "1",
                                              start_end_series_mixed_case)
    _verify_ten_eleven_rates(result)
    result = _event_analysis_by_activity_case(ground_truth_mixed_case,
                                              detected_mixed_case,
                                              "*",
                                              "2",
                                              start_end_series_mixed_case)
    _verify_ten_eighteen_rates(result)
    result = _event_analysis_by_activity_case(ground_truth_mixed_case,
                                              detected_mixed_case,
                                              "*",
                                              "*",
                                              start_end_series_mixed_case)
    assert isinstance(result, EventAnalysis)
    assert result.dr == 0.0
    assert result.fr == 0.5
    assert result.fmr == 0.0
    assert result.mr == 0.0
    assert result.cr_gt == 0.5
    assert result.mdr == 0.0
    assert result.fdr == 0.6
    assert result.idr == 0.1
    assert result.cr_det == 0.3


def test_context_ea_1_a() -> None:
    """Test EA metrics that are exposed in the context class for trace 1, activity A."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis(activity_name="Activity A", case_id="ExampleTrace1")
    assert isinstance(res, EventAnalysis)
    assert res.d == 0
    assert res.f == 1
    assert res.fm == 0
    assert res.m == 0
    assert res.c == 2
    assert res.md == 0
    assert res.fmd == 0
    assert res.fd == 2
    assert res.id == 0


def test_context_ea_1_b() -> None:
    """Test EA metrics that are exposed in the context class for trace 1, activity B."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis(activity_name="Activity B", case_id="ExampleTrace1")
    assert isinstance(res, EventAnalysis)
    assert res.d == 0
    assert res.f == 0
    assert res.fm == 0
    assert res.m == 2
    assert res.c == 0
    assert res.md == 1
    assert res.fmd == 0
    assert res.fd == 0
    assert res.id == 0


def test_context_ea_1_c() -> None:
    """Test EA metrics that are exposed in the context class for trace 1, activity C."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis(activity_name="Activity C", case_id="ExampleTrace1")
    assert isinstance(res, EventAnalysis)
    assert res.d == 1
    assert res.f == 0
    assert res.fm == 0
    assert res.m == 0
    assert res.c == 2
    assert res.md == 0
    assert res.fmd == 0
    assert res.fd == 0
    assert res.id == 0


def test_context_ea_2_a() -> None:
    """Test EA metrics that are exposed in the context class for trace 2, activity A."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis(activity_name="Activity A", case_id="ExampleTrace2")
    assert isinstance(res, EventAnalysis)
    assert res.d == 1
    assert res.f == 0
    assert res.fm == 0
    assert res.m == 0
    assert res.c == 0
    assert res.md == 0
    assert res.fmd == 0
    assert res.fd == 0
    assert res.id == 1


def test_context_ea_2_b() -> None:
    """Test EA metrics that are exposed in the context class for trace 2, activity B."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis(activity_name="Activity B", case_id="ExampleTrace2")
    assert isinstance(res, EventAnalysis)
    assert res.d == 0
    assert res.f == 0
    assert res.fm == 1
    assert res.m == 1
    assert res.c == 0
    assert res.md == 0
    assert res.fmd == 1
    assert res.fd == 1
    assert res.id == 0


def test_context_ea_rates_1_a() -> None:
    """Test EA rates that are exposed in the context class for trace 1, activity A."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis(activity_name="Activity A", case_id="ExampleTrace1")
    assert isinstance(res, EventAnalysis)
    assert res.dr == 0.0
    assert res.fr == 0.5
    assert res.fmr == 0.0
    assert res.mr == 0.0
    assert res.cr_gt == 0.5
    assert res.cr_det == round(1 / 3, 4)
    assert res.mdr == 0.0
    assert res.fmdr == 0.0
    assert res.fdr == round(2 / 3, 4)
    assert res.idr == 0.0


def test_context_ea_rates_1_b() -> None:
    """Test EA rates that are exposed in the context class for trace 1, activity B."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis(activity_name="Activity B", case_id="ExampleTrace1")
    assert isinstance(res, EventAnalysis)
    assert res.dr == 0.0
    assert res.fr == 0.0
    assert res.fmr == 0.0
    assert res.mr == 1.0
    assert res.cr_gt == 0.0
    assert res.cr_det == 0.0
    assert res.mdr == 1.0
    assert res.fmdr == 0.0
    assert res.fdr == 0.0
    assert res.idr == 0.0


def test_context_ea_rates_1_c() -> None:
    """Test EA rates that are exposed in the context class for trace 1, activity C."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis(activity_name="Activity C", case_id="ExampleTrace1")
    assert isinstance(res, EventAnalysis)
    assert res.dr == 0.5
    assert res.fr == 0.0
    assert res.fmr == 0.0
    assert res.mr == 0.0
    assert res.cr_gt == 0.5
    assert res.cr_det == 1.0
    assert res.mdr == 0.0
    assert res.fmdr == 0.0
    assert res.fdr == 0.0
    assert res.idr == 0.0


def test_context_ea_rates_2_a() -> None:
    """Test EA rates that are exposed in the context class for trace 2, activity A."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis(activity_name="Activity A", case_id="ExampleTrace2")
    assert isinstance(res, EventAnalysis)
    assert res.dr == 1.0
    assert res.fr == 0.0
    assert res.fmr == 0.0
    assert res.mr == 0.0
    assert res.cr_gt == 0.0
    assert res.cr_det == 0.0
    assert res.mdr == 0.0
    assert res.fmdr == 0.0
    assert res.fdr == 0.0
    assert res.idr == 1.0


def test_context_ea_rates_2_b() -> None:
    """Test EA rates that are exposed in the context class for trace 2, activity B."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis(activity_name="Activity B", case_id="ExampleTrace2")
    assert isinstance(res, EventAnalysis)
    assert res.dr == 0.0
    assert res.fr == 0.0
    assert res.fmr == 0.5
    assert res.mr == 0.5
    assert res.cr_gt == 0.0
    assert res.cr_det == 0.0
    assert res.mdr == 0.0
    assert res.fmdr == 0.5
    assert res.fdr == 0.5
    assert res.idr == 0.0


def test_context_ea_by_activity() -> None:
    """Test EA by activity that is exposed in the context class."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis(activity_name="Activity A")
    assert isinstance(res, EventAnalysis)
    assert res.d == round((0 + 1) / 2, 4)
    assert res.f == round((1 + 0) / 2, 4)
    assert res.fm == round((0 + 0) / 2, 4)
    assert res.m == round((0 + 0) / 2, 4)
    assert res.c == round((2 + 0) / 2, 4)
    assert res.md == round((0 + 0) / 2, 4)
    assert res.fmd == round((0 + 0) / 2, 4)
    assert res.fd == round((2 + 0) / 2, 4)
    assert res.id == round((0 + 1) / 2, 4)
    assert res.total_gt_events == 1.5
    assert res.total_det_events == 2
    res = context.event_analysis(activity_name="Activity B")
    assert isinstance(res, EventAnalysis)
    assert res.d == round((0 + 0) / 2, 4)
    assert res.f == round((0 + 0) / 2, 4)
    assert res.fm == round((0 + 1) / 2, 4)
    assert res.m == round((2 + 1) / 2, 4)
    assert res.c == round((0 + 0) / 2, 4)
    assert res.md == round((0 + 1) / 2, 4)
    assert res.fmd == round((0 + 1) / 2, 4)
    assert res.fd == round((0 + 1) / 2, 4)
    assert res.id == round((0 + 0) / 2, 4)
    assert res.total_gt_events == 2
    assert res.total_det_events == 1.5
    res = context.event_analysis(activity_name="Activity C")
    assert isinstance(res, EventAnalysis)
    assert res.d == 1
    assert res.f == 0
    assert res.fm == 0
    assert res.m == 0
    assert res.c == 2
    assert res.md == 0
    assert res.fmd == 0
    assert res.fd == 0
    assert res.id == 0
    assert res.total_gt_events == 2
    assert res.total_det_events == 1


def test_context_ea_by_activity_case() -> None:
    """Test EA by activity and case that is exposed in the context class."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis()
    assert isinstance(res, EventAnalysis)
    assert res.d == round((0 + 0 + 1 + 1 + 0) / 5, 4)
    assert res.f == round((1 + 0 + 0 + 0 + 0) / 5, 4)
    assert res.fm == round((0 + 0 + 0 + 0 + 1) / 5, 4)
    assert res.m == round((0 + 2 + 0 + 0 + 1) / 5, 4)
    assert res.c == round((2 + 0 + 2 + 0 + 0) / 5, 4)
    assert res.md == round((0 + 1 + 0 + 0 + 0) / 5, 4)
    assert res.fmd == round((0 + 0 + 0 + 0 + 1) / 5, 4)
    assert res.fd == round((2 + 0 + 0 + 0 + 1) / 5, 4)
    assert res.id == round((0 + 0 + 0 + 0 + 1) / 5, 4)
    assert res.total_gt_events == 1.8
    assert res.total_det_events == 1.6


def test_context_ea_rates_by_activity() -> None:
    """Test EA rates by activity that is exposed in the context class, aggregated by activity."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis(activity_name="Activity A")
    assert isinstance(res, EventAnalysis)
    assert res.dr == 0.3333
    assert res.fr == 0.3333
    assert res.fmr == 0
    assert res.mr == 0
    assert res.cr_gt == 0.3333
    assert res.cr_det == 0.25
    assert res.mdr == 0
    assert res.fmdr == 0
    assert res.fdr == 0.5
    assert res.idr == 0.25
    res = context.event_analysis(activity_name="Activity B")
    assert isinstance(res, EventAnalysis)
    assert res.dr == 0
    assert res.fr == 0
    assert res.fmr == 0.25
    assert res.mr == 0.75
    assert res.cr_gt == 0
    assert res.cr_det == 0
    assert res.mdr == 0.3333
    assert res.fmdr == 0.3333
    assert res.fdr == 0.3333
    assert res.idr == 0
    res = context.event_analysis(activity_name="Activity C")
    assert isinstance(res, EventAnalysis)
    assert res.dr == 0.5
    assert res.fr == 0.0
    assert res.fmr == 0.0
    assert res.mr == 0.0
    assert res.cr_gt == 0.5
    assert res.cr_det == 1.0
    assert res.mdr == 0.0
    assert res.fmdr == 0.0
    assert res.fdr == 0.0
    assert res.idr == 0.0


def test_context_ea_rates_by_activity_case() -> None:
    """Test EA rates by activity and case that is exposed in the context class,
    aggregated by activity and case."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis()
    assert isinstance(res, EventAnalysis)
    assert res.dr == 0.2222
    assert res.fr == 0.1111
    assert res.fmr == 0.1111
    assert res.mr == 0.3333
    assert res.cr_gt == 0.2222
    assert res.cr_det == 0.25
    assert res.mdr == 0.125
    assert res.fmdr == 0.125
    assert res.fdr == 0.375
    assert res.idr == 0.125


def test_ea_recall_precision_f1() -> None:
    """Test EA rates by activity and case that is exposed in the context class,
    aggregated by activity and case."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.event_analysis()
    assert isinstance(res, EventAnalysis)
    assert 0 <= res.recall <= 1
    assert 0 <= res.precision <= 1
    assert 0 <= res.f1 <= 1



"""This module contains tests for the aqudem utils module."""
import os
import pytest
import aqudem


def test_xes_check_missing_sampling_freq() -> None:
    """Test the XESMissingSamplingFreqError exception."""
    with pytest.raises(aqudem.utils.XESSamplingFreqError):
        aqudem.Context(os.path.join("tests", "resources", "ground_truth_missingsamplingfreq.xes"),
                       os.path.join("tests", "resources", "detected_missingsamplingfreq.xes"))


def test_xes_missing_lifecycle_transition() -> None:
    """Test the XESIncorrectLifecycleTransitionError exception."""
    with pytest.raises(aqudem.utils.XESIncorrectLifecycleTransitionError):
        aqudem.Context(os.path.join("tests", "resources", "ground_truth_missinglifecycle.xes"),
                       os.path.join("tests", "resources", "detected_missinglifecycle.xes"))


def test_xes_wrong_lifecycle_transition() -> None:
    """Test the XESIncorrectLifecycleTransitionError exception."""
    with pytest.raises(aqudem.utils.XESIncorrectLifecycleTransitionError):
        aqudem.Context(os.path.join("tests", "resources", "ground_truth_wronglifecycle.xes"),
                       os.path.join("tests", "resources", "detected_wronglifecycle.xes"))


def test_xes_toomanycomplete_lifecycle_transition() -> None:
    """Test the XESIncorrectLifecycleTransitionError exception."""
    with pytest.raises(aqudem.utils.XESIncorrectLifecycleTransitionError):
        aqudem.Context(os.path.join("tests",
                                    "resources",
                                    "ground_truth_toomanycompletelifecycle.xes"),
                       os.path.join("tests", "resources", "detected_toomanycompletelifecycle.xes"))


def test_xes_missing_timestamp() -> None:
    """Test the XESMissingTimestamp exception."""
    with pytest.raises(aqudem.utils.XESMissingTimestamp):
        aqudem.Context(os.path.join("tests", "resources", "ground_truth_missingtimestamp.xes"),
                       os.path.join("tests", "resources", "detected_missingtimestamp.xes"))


def test_xes_missing_activity_name() -> None:
    """Test the XESMissingActivityName exception."""
    with pytest.raises(aqudem.utils.XESMissingActivityName):
        aqudem.Context(os.path.join("tests", "resources", "ground_truth_missingactivityname.xes"),
                       os.path.join("tests", "resources", "detected_missingactivityname.xes"))


def test_xes_missing_trace_name() -> None:
    """Test the XESMissingTraceNameAttribute exception."""
    with pytest.raises(aqudem.utils.XESMissingTraceNameAttribute):
        aqudem.Context(os.path.join("tests", "resources", "ground_truth_missingtracename.xes"),
                       os.path.join("tests", "resources", "detected_missingtracename.xes"))


def test_initialization_error_case_missing() -> None:
    """Test the initialization of the context object with case mismatch/missing case."""
    with pytest.raises(ValueError) as e:
        aqudem.Context(os.path.join("tests", "resources", "ground_truth_no_matching_case.xes"),
                       os.path.join("tests", "resources", "detected_no_matching_case.xes"))
    assert "Currently, all case IDs must be in both logs." in str(e.value)


def test_initialization_inconsistent_samping_freq() -> None:
    """Test the initialization of the context object with inconsistent sampling frequency."""
    with pytest.raises(aqudem.utils.XESSamplingFreqError):
        aqudem.Context(
            os.path.join("tests", "resources", "ground_truth_inconsistent_sampling_freq.xes"),
            os.path.join("tests", "resources", "detected_inconsistent_sampling_freq.xes"))



""" Testing the aqudem package end-to-end. """
import os
from collections import Counter
from typing import Set

import pytest
import static_frame as sf
import aqudem

ACTIVIES_GT_03_20 = ['OV Burn', 'Sorting Machine sort', 'WT Transport', 'Calibrate VGR',
               'Get Workpiece from Pickup Station', 'Pickup and transport to sink',
               'Pickup and transport to Oven', 'Unload from HBW', 'Store Workpiece in HBW',
               'Calibrate HBW', 'Start Milling Machine', 'Move to DPS']
ACTIVITIES_DET_03_20 = ['WT Transport', 'Get Workpiece from Pickup Station',
                  'Store Workpiece in HBW', 'Calibrate HBW', 'Read Color', 'Move to DPS']
CASES_GT_03_20 = ['case1']
CASES_DET_03_20 = ['case1']

def get_set_all_activities(context: aqudem.Context) -> Set[str]:
    return set(context.activity_names["ground_truth"] + context.activity_names["detected"])

def get_set_all_cases(context: aqudem.Context) -> Set[str]:
    return set(context.case_ids["ground_truth"] + context.case_ids["detected"])

def act_only_in_one_log(context: aqudem.Context) -> Set[str]:
    return set(context.activity_names["ground_truth"]) ^ set(context.activity_names["detected"])

def case_only_in_one_log(context: aqudem.Context) -> Set[str]:
    return set(context.case_ids["ground_truth"]) ^ set(context.case_ids["detected"])


def _validate_two_set(two_set: aqudem.TwoSet) -> None:
    assert isinstance(two_set, aqudem.TwoSet)
    for key in ["tp", "tn", "d", "f", "ua", "uo", "i", "m", "oa", "oo",
                "p", "n", "t",
                "tpr", "tnr", "dr", "fr", "uar", "uor", "ir", "mr", "oar", "oor"]:
        assert isinstance(getattr(two_set, key), (int, float))
    assert two_set.tp + two_set.d + two_set.f + two_set.ua + two_set.uo == two_set.p
    assert two_set.tn + two_set.i + two_set.m + two_set.oa + two_set.oo == two_set.n
    assert round(two_set.p + two_set.n, 2) == round(two_set.t, 2)
    assert round(two_set.tpr + two_set.dr + two_set.fr + two_set.uar + two_set.uor, 4) - 1 < 0.001
    assert round(two_set.tnr + two_set.ir + two_set.mr + two_set.oar + two_set.oor, 4) - 1 < 0.001


def _validate_two_set_zero(two_set: aqudem.TwoSet) -> None:
    assert isinstance(two_set, aqudem.TwoSet)
    for key in ["tp", "f", "ua", "uo", "m", "oa", "oo",
                "tpr", "fr", "uar", "uor", "mr", "oar", "oor"]:
        assert getattr(two_set, key) == 0
    assert two_set.d > 0 or two_set.i > 0
    assert not (two_set.d > 0 and two_set.i > 0)
    assert two_set.dr > 0 or two_set.ir > 0
    assert not (two_set.dr > 0 and two_set.ir > 0)


def _validate_event_analysis(ea: aqudem.EventAnalysis) -> None:
    assert isinstance(ea, aqudem.EventAnalysis)
    for key in ["d", "f", "fm", "m", "c", "md", "fmd", "fd", "id",
                "total_gt_events", "total_det_events", "correct_events_per_log",
                "dr", "fr", "fmr", "mr", "cr_gt", "cr_det", "mdr", "fmdr", "fdr", "idr"]:
        assert isinstance(getattr(ea, key), (int, float))
    assert (ea.dr + ea.fr + ea.fmr + ea.mr + ea.cr_gt) - 1 < 0.001
    assert (ea.mdr + ea.fmdr + ea.fdr + ea.idr + ea.cr_det) - 1 < 0.001
    assert (ea.d + ea.f + ea.fm + ea.m + (ea.c / 2)) - ea.total_gt_events < 0.001
    assert (ea.md + ea.fd + ea.fmd + ea.id + (ea.c / 2)) - ea.total_det_events < 0.001


def _validate_event_analysis_zero(ea: aqudem.EventAnalysis) -> None:
    assert isinstance(ea, aqudem.EventAnalysis)
    for key in ["f", "fm", "m", "c", "md", "fmd", "fd",
                "correct_events_per_log",
                "fr", "fmr", "mr", "cr_gt", "cr_det", "mdr", "fmdr", "fdr"]:
        assert getattr(ea, key) == 0
    assert ea.d > 0 or ea.id > 0
    assert not (ea.d > 0 and ea.id > 0)
    assert ea.dr > 0 or ea.idr > 0
    assert not (ea.dr > 0 and ea.idr > 0)
    assert ea.total_gt_events > 0 or ea.total_det_events > 0
    assert not (ea.total_gt_events > 0 and ea.total_det_events > 0)


@pytest.fixture(scope="module", name='context')
def fixture_context(request: pytest.FixtureRequest) -> aqudem.Context:
    file1, file2 = request.param
    return aqudem.Context(os.path.join("tests", "resources", "experiment-logs", file1),
                          os.path.join("tests", "resources", "experiment-logs", file2))


@pytest.mark.parametrize("context", [
    ("23-03-20_gt_camunda.xes", "23-03-20_det_firstlastlowlevel.xes")
], indirect=True)
def test_context_properties_3_20(context: aqudem.Context) -> None:
    assert isinstance(context.ground_truth, sf.FrameHE)
    assert isinstance(context.detected, sf.FrameHE)
    assert context.ground_truth.shape[0] <= 204
    assert context.ground_truth.shape[1] == 5
    for column in ["case:concept:name", "concept:name",
                   "lifecycle:transition", "time:timestamp", "concept:instance"]:
        assert column in context.ground_truth.columns
    assert context.detected.shape[0] <= 78
    assert context.detected.shape[1] == 5
    for column in ["case:concept:name", "concept:name", "lifecycle:transition",
                   "time:timestamp", "case:sampling_freq"]:
        assert column in context.detected.columns
    assert isinstance(context.activity_names, dict)
    assert isinstance(context.case_ids, dict)
    assert Counter(context.activity_names["ground_truth"]) == Counter(ACTIVIES_GT_03_20)
    assert Counter(context.activity_names["detected"]) == Counter(ACTIVITIES_DET_03_20)
    assert Counter(context.case_ids["ground_truth"]) == Counter(CASES_GT_03_20)
    assert Counter(context.case_ids["detected"]) == Counter(CASES_DET_03_20)


@pytest.mark.parametrize("context", [
    ("23-03-20_gt_camunda.xes", "23-03-20_det_firstlastlowlevel.xes"),
    ("23_01_30_gt_camunda.xes", "23_01_30_det_firstlastlowlevel.xes"),
    ("23_02_06_gt_camunda.xes", "23_02_06_det_firstlastlowlevel.xes"),
    ("23_04_11_gt_camunda.xes", "23_04_11_det_firstlastlowlevel.xes"),
    ("23-01-02-03-04_gt_comb.xes", "23-01-02-03-04_det_comb.xes"),
], indirect=True)
def test_context_properties(context: aqudem.Context) -> None:
    assert isinstance(context.ground_truth, sf.FrameHE)
    assert isinstance(context.detected, sf.FrameHE)
    for column in ["case:concept:name", "concept:name",
                   "lifecycle:transition", "time:timestamp", "concept:instance"]:
        assert column in context.ground_truth.columns
    for column in ["case:concept:name", "concept:name", "lifecycle:transition",
                   "time:timestamp", "case:sampling_freq"]:
        assert column in context.detected.columns
    assert isinstance(context.activity_names, dict)
    assert isinstance(context.case_ids, dict)

@pytest.mark.parametrize("context", [
    ("23-03-20_gt_camunda.xes", "23-03-20_det_firstlastlowlevel.xes"),
    ("23_01_30_gt_camunda.xes", "23_01_30_det_firstlastlowlevel.xes"),
    ("23_02_06_gt_camunda.xes", "23_02_06_det_firstlastlowlevel.xes"),
    ("23_04_11_gt_camunda.xes", "23_04_11_det_firstlastlowlevel.xes"),
    ("23-01-02-03-04_gt_comb.xes", "23-01-02-03-04_det_comb.xes"),

], indirect=True)
def test_cross_correlation(context: aqudem.Context) -> None:
    cross_correlation = context.cross_correlation()
    assert isinstance(cross_correlation, tuple)
    assert isinstance(cross_correlation[0], (float, int))
    assert isinstance(cross_correlation[1], float)

    for act in get_set_all_activities(context):
        cross_correlation_act = context.cross_correlation(activity_name=act)
        assert isinstance(cross_correlation_act, tuple)
        assert isinstance(cross_correlation_act[0], (float, int))
        assert isinstance(cross_correlation_act[1], float)

    for cas in get_set_all_cases(context):
        cross_correlation_case = context.cross_correlation(case_id=cas)
        assert isinstance(cross_correlation_case, tuple)
        assert isinstance(cross_correlation_case[0], (float, int))
        assert isinstance(cross_correlation_case[1], float)

    for act in get_set_all_activities(context):
        for cas in get_set_all_cases(context):
            try:
                cross_correlation_act_case = context.cross_correlation(activity_name=act,
                                                                       case_id=cas)
                assert isinstance(cross_correlation_act_case, tuple)
                assert isinstance(cross_correlation_act_case[0], (float, int))
                assert isinstance(cross_correlation_act_case[1], float)
            except ValueError as e:
                assert "No metrics could be calculated for this combination" in str(e)

    # for activities that are only in one log, make sure that they are logically ZERO
    for act in act_only_in_one_log(context):
        cross_correlation_act = context.cross_correlation(activity_name=act)
        assert cross_correlation_act[0] == 0
        assert cross_correlation_act[1] == 0


@pytest.mark.parametrize("context", [
    ("23-03-20_gt_camunda.xes", "23-03-20_det_firstlastlowlevel.xes"),
    ("23_01_30_gt_camunda.xes", "23_01_30_det_firstlastlowlevel.xes"),
    ("23_02_06_gt_camunda.xes", "23_02_06_det_firstlastlowlevel.xes"),
    ("23_04_11_gt_camunda.xes", "23_04_11_det_firstlastlowlevel.xes"),
    ("23-01-02-03-04_gt_comb.xes", "23-01-02-03-04_det_comb.xes"),

], indirect=True)
def test_two_set(context: aqudem.Context) -> None:
    two_set = context.two_set()
    _validate_two_set(two_set)

    for act in get_set_all_activities(context):
        two_set_act = context.two_set(activity_name=act)
        _validate_two_set(two_set_act)

    for cas in get_set_all_cases(context):
        two_set_case = context.two_set(case_id=cas)
        _validate_two_set(two_set_case)

    for act in get_set_all_activities(context):
        for cas in get_set_all_cases(context):
            try:
                two_set_act_case = context.two_set(activity_name=act, case_id=cas)
                _validate_two_set(two_set_act_case)
            except ValueError as e:
                assert "No metrics could be calculated for this combination" in str(e)

    for act in act_only_in_one_log(context):
        two_set_act = context.two_set(activity_name=act)
        _validate_two_set_zero(two_set_act)


@pytest.mark.parametrize("context", [
    ("23-03-20_gt_camunda.xes", "23-03-20_det_firstlastlowlevel.xes"),
    ("23_01_30_gt_camunda.xes", "23_01_30_det_firstlastlowlevel.xes"),
    ("23_02_06_gt_camunda.xes", "23_02_06_det_firstlastlowlevel.xes"),
    ("23_04_11_gt_camunda.xes", "23_04_11_det_firstlastlowlevel.xes"),
    ("23-01-02-03-04_gt_comb.xes", "23-01-02-03-04_det_comb.xes"),

], indirect=True)
def test_event_analysis(context: aqudem.Context) -> None:
    ea = context.event_analysis()
    _validate_event_analysis(ea)

    for act in get_set_all_activities(context):
        ea_act = context.event_analysis(activity_name=act)
        _validate_event_analysis(ea_act)

    for cas in get_set_all_cases(context):
        ea_case = context.event_analysis(case_id=cas)
        _validate_event_analysis(ea_case)

    for act in get_set_all_activities(context):
        for cas in get_set_all_cases(context):
            try:
                ea_act_case = context.event_analysis(activity_name=act, case_id=cas)
                _validate_event_analysis(ea_act_case)
            except ValueError as e:
                assert "No metrics could be calculated for this combination" in str(e)

    for act in act_only_in_one_log(context):
        ea_act = context.event_analysis(activity_name=act)
        _validate_event_analysis_zero(ea_act)


@pytest.mark.parametrize("context", [
    ("23-03-20_gt_camunda.xes", "23-03-20_det_firstlastlowlevel.xes"),
    ("23_01_30_gt_camunda.xes", "23_01_30_det_firstlastlowlevel.xes"),
    ("23_02_06_gt_camunda.xes", "23_02_06_det_firstlastlowlevel.xes"),
    ("23_04_11_gt_camunda.xes", "23_04_11_det_firstlastlowlevel.xes"),
    ("23-01-02-03-04_gt_comb.xes", "23-01-02-03-04_det_comb.xes"),

], indirect=True)
def test_damerau_levenshtein_distance(context: aqudem.Context) -> None:
    dld = context.damerau_levenshtein_distance()
    assert isinstance(dld[0], (int, float))
    assert isinstance(dld[1], float)
    assert dld[1] <= 1

    for cas in get_set_all_cases(context):
        dld_case = context.damerau_levenshtein_distance(case_id=cas)
        assert isinstance(dld_case[0], (int, float))
        assert isinstance(dld_case[1], float)
        assert dld_case[1] <= 1


@pytest.mark.parametrize("context", [
    ("23-03-20_gt_camunda.xes", "23-03-20_det_firstlastlowlevel.xes"),
    ("23_01_30_gt_camunda.xes", "23_01_30_det_firstlastlowlevel.xes"),
    ("23_02_06_gt_camunda.xes", "23_02_06_det_firstlastlowlevel.xes"),
    ("23_04_11_gt_camunda.xes", "23_04_11_det_firstlastlowlevel.xes"),
    ("23-01-02-03-04_gt_comb.xes", "23-01-02-03-04_det_comb.xes"),

], indirect=True)
def test_levenshtein_distance(context: aqudem.Context) -> None:
    ld = context.levenshtein_distance()
    assert isinstance(ld[0], (int, float))
    assert isinstance(ld[1], float)
    assert ld[1] <= 1

    for cas in get_set_all_cases(context):
        ld_case = context.levenshtein_distance(case_id=cas)
        assert isinstance(ld_case[0], (int, float))
        assert isinstance(ld_case[1], float)
        assert ld_case[1] <= 1


@pytest.mark.parametrize("context", [
    ("23-03-20_gt_camunda.xes", "23-03-20_det_firstlastlowlevel.xes")
], indirect=True)
def test_different_ordering_xes_same_result(context: aqudem.Context) -> None:
    context2 = aqudem.Context(os.path.join("tests",
                                           "resources",
                                           "23-03-20_gt_cam_ooo.xes"),
                              os.path.join("tests",
                                           "resources",
                                           "23-03-20_det_firstlastlowlevel_ooo.xes"))
    assert context.activity_names == context2.activity_names
    assert context.case_ids == context2.case_ids
    assert context.cross_correlation() == context2.cross_correlation()
    assert context.two_set() == context2.two_set()
    assert context.event_analysis() == context2.event_analysis()
    assert context.damerau_levenshtein_distance() == context2.damerau_levenshtein_distance()
    assert context.levenshtein_distance() == context2.levenshtein_distance()


@pytest.mark.parametrize("context", [
    ("23-03-20_gt_camunda.xes", "23-03-20_det_firstlastlowlevel.xes")
], indirect=True)
def test_wrong_case_or_activity_raises_exception(context: aqudem.Context) -> None:
    with pytest.raises(ValueError):
        context.two_set(activity_name="wrong_activity_name")
    with pytest.raises(ValueError):
        context.two_set(case_id="wrong_case_id")


@pytest.mark.parametrize("context", [
    ("23_04_11_gt_camunda.xes", "23_04_11_det_firstlastlowlevel.xes")
], indirect=True)
def test_no_duration_activities_removed(context: aqudem.Context) -> None:
    """ Test that activities with no duration are removed. """
    with pytest.raises(ValueError) as e:
        context.cross_correlation(activity_name="Read Color")
    assert "The activity name 'Read Color' is not in the event logs." in str(e)

def test_empty_det_log() -> None:
    """ Test that empty detected raises a ValueError on context creation. """
    with pytest.raises(ValueError) as e:
        aqudem.Context(os.path.join("tests", "resources", "experiment-logs", "no_det_gt.xes"),
                       os.path.join("tests", "resources", "experiment-logs", "no_det_det.xes"))
    assert "empty" in str(e)



"""This module contains tests for the aqudem utils module."""
from datetime import datetime
import numpy as np
import static_frame as sf
import pytest
from aqudem import ward_helper
from aqudem.ward_helper import _is_during_activity_exec
from .mocks.logs import detected_ten_eleven, ground_truth_ten_eleven, start_end_series_ten_eleven


def test_generate_eight_type_with_valid_input() -> None:
    assert ward_helper._generate_eight_type("TP", "FP", "TN") == "Oo"


def test_generate_eight_type_with_invalid_segment_type() -> None:
    with pytest.raises(ValueError):
        ward_helper._generate_eight_type("Invalid", "FP", "TN")


def test_generate_eight_type_with_invalid_curr_type() -> None:
    with pytest.raises(ValueError):
        ward_helper._generate_eight_type("TP", "Invalid", "TN")


def test_generate_eight_type_with_invalid_prev_type() -> None:
    with pytest.raises(ValueError):
        ward_helper._generate_eight_type("NaN", "FP", "TN")


def test_generate_eight_type_with_invalid_next_type() -> None:
    with pytest.raises(ValueError):
        ward_helper._generate_eight_type("TP", "FP", "Invalid")


def test_generate_segment_scores_with_valid_input() -> None:
    ground_truth = sf.FrameHE.from_dict({
        "lifecycle:transition": ["start", "complete", "start", "complete", "start", "complete"],
        "time:timestamp": [datetime(2021, 1, 1, 10, 0), datetime(2021, 1, 1, 12, 0),
                           datetime(2021, 1, 1, 13, 0), datetime(2021, 1, 1, 15, 0),
                           datetime(2021, 1, 1, 16, 0), datetime(2021, 1, 1, 18, 0)]
    }, dtypes={"lifecycle:transition": str, "time:timestamp": np.datetime64})
    detected = sf.FrameHE.from_dict({
        "lifecycle:transition": ["start", "complete", "start", "complete", "start", "complete",
                                 "start", "complete", "start", "complete"],
        "time:timestamp": [datetime(2021, 1, 1, 10, 15), datetime(2021, 1, 1, 11, 0),
                           datetime(2021, 1, 1, 11, 30), datetime(2021, 1, 1, 12, 30),
                           datetime(2021, 1, 1, 13, 0), datetime(2021, 1, 1, 14, 0),
                           datetime(2021, 1, 1, 14, 30), datetime(2021, 1, 1, 15, 30),
                           datetime(2021, 1, 1, 15, 45), datetime(2021, 1, 1, 17, 0)]
    }, dtypes={"lifecycle:transition": str, "time:timestamp": np.datetime64})
    result = ward_helper._generate_segment_scores(ground_truth,
                                                  detected,
                                                  datetime(2021, 1, 1, 10, 0),
                                                  datetime(2021, 1, 1, 18, 0))
    assert len(result) != 0
    assert "start" in result.columns
    assert "end" in result.columns
    assert "type" in result.columns

    assert result.iloc[0]["type"] == "Ua"
    assert result.iloc[0]["start"] == datetime(2021, 1, 1, 10, 0)
    assert result.iloc[0]["end"] == datetime(2021, 1, 1, 10, 15)

    assert result.iloc[1]["type"] == "TP"
    assert result.iloc[1]["start"] == datetime(2021, 1, 1, 10, 15)
    assert result.iloc[1]["end"] == datetime(2021, 1, 1, 11, 0)

    assert result.iloc[2]["type"] == "F"
    assert result.iloc[2]["start"] == datetime(2021, 1, 1, 11, 0)
    assert result.iloc[2]["end"] == datetime(2021, 1, 1, 11, 30)

    assert result.iloc[3]["type"] == "TP"
    assert result.iloc[3]["start"] == datetime(2021, 1, 1, 11, 30)
    assert result.iloc[3]["end"] == datetime(2021, 1, 1, 12, 0)

    assert result.iloc[4]["type"] == "Oo"
    assert result.iloc[4]["start"] == datetime(2021, 1, 1, 12, 0)
    assert result.iloc[4]["end"] == datetime(2021, 1, 1, 12, 30)

    assert result.iloc[5]["type"] == "TN"
    assert result.iloc[5]["start"] == datetime(2021, 1, 1, 12, 30)
    assert result.iloc[5]["end"] == datetime(2021, 1, 1, 13, 0)

    assert result.iloc[6]["type"] == "TP"
    assert result.iloc[6]["start"] == datetime(2021, 1, 1, 13, 0)
    assert result.iloc[6]["end"] == datetime(2021, 1, 1, 14, 0)

    assert result.iloc[7]["type"] == "F"
    assert result.iloc[7]["start"] == datetime(2021, 1, 1, 14, 0)
    assert result.iloc[7]["end"] == datetime(2021, 1, 1, 14, 30)

    assert result.iloc[8]["type"] == "TP"
    assert result.iloc[8]["start"] == datetime(2021, 1, 1, 14, 30)
    assert result.iloc[8]["end"] == datetime(2021, 1, 1, 15, 0)

    assert result.iloc[9]["type"] == "Oo"
    assert result.iloc[9]["start"] == datetime(2021, 1, 1, 15, 0)
    assert result.iloc[9]["end"] == datetime(2021, 1, 1, 15, 30)

    assert result.iloc[10]["type"] == "TN"
    assert result.iloc[10]["start"] == datetime(2021, 1, 1, 15, 30)
    assert result.iloc[10]["end"] == datetime(2021, 1, 1, 15, 45)

    assert result.iloc[11]["type"] == "Oa"
    assert result.iloc[11]["start"] == datetime(2021, 1, 1, 15, 45)
    assert result.iloc[11]["end"] == datetime(2021, 1, 1, 16, 0)

    assert result.iloc[12]["type"] == "TP"
    assert result.iloc[12]["start"] == datetime(2021, 1, 1, 16, 0)
    assert result.iloc[12]["end"] == datetime(2021, 1, 1, 17, 0)

    assert result.iloc[13]["type"] == "Uo"
    assert result.iloc[13]["start"] == datetime(2021, 1, 1, 17, 0)
    assert result.iloc[13]["end"] == datetime(2021, 1, 1, 18, 0)


# pylint: disable=too-many-statements, unsubscriptable-object
def test_generate_segment_scores_with_valid_input_2() -> None:
    result = ward_helper._generate_segment_scores(ground_truth_ten_eleven,
                                                  detected_ten_eleven,
                                                  start_end_series_ten_eleven["1"][0],
                                                  start_end_series_ten_eleven["1"][1])
    assert len(result) != 0
    assert "start" in result.columns
    assert "end" in result.columns
    assert "type" in result.columns

    assert result.iloc[0]["type"] == "Oa"
    assert result.iloc[0]["start"] == datetime(2021, 1, 1, 10, 0)
    assert result.iloc[0]["end"] == datetime(2021, 1, 1, 10, 5)

    assert result.iloc[1]["type"] == "TP"
    assert result.iloc[1]["start"] == datetime(2021, 1, 1, 10, 5)
    assert result.iloc[1]["end"] == datetime(2021, 1, 1, 10, 10)

    assert result.iloc[2]["type"] == "F"
    assert result.iloc[2]["start"] == datetime(2021, 1, 1, 10, 10)
    assert result.iloc[2]["end"] == datetime(2021, 1, 1, 10, 15)

    assert result.iloc[3]["type"] == "TP"
    assert result.iloc[3]["start"] == datetime(2021, 1, 1, 10, 15)
    assert result.iloc[3]["end"] == datetime(2021, 1, 1, 10, 20)

    assert result.iloc[4]["type"] == "Uo"
    assert result.iloc[4]["start"] == datetime(2021, 1, 1, 10, 20)
    assert result.iloc[4]["end"] == datetime(2021, 1, 1, 10, 25)

    assert result.iloc[5]["type"] == "TN"
    assert result.iloc[5]["start"] == datetime(2021, 1, 1, 10, 25)
    assert result.iloc[5]["end"] == datetime(2021, 1, 1, 10, 30)

    assert result.iloc[6]["type"] == "Oa"
    assert result.iloc[6]["start"] == datetime(2021, 1, 1, 10, 30)
    assert result.iloc[6]["end"] == datetime(2021, 1, 1, 10, 35)

    assert result.iloc[7]["type"] == "TP"
    assert result.iloc[7]["start"] == datetime(2021, 1, 1, 10, 35)
    assert result.iloc[7]["end"] == datetime(2021, 1, 1, 10, 38)

    assert result.iloc[8]["type"] == "Oo"
    assert result.iloc[8]["start"] == datetime(2021, 1, 1, 10, 38)
    assert result.iloc[8]["end"] == datetime(2021, 1, 1, 10, 40)

    assert result.iloc[9]["type"] == "TN"
    assert result.iloc[9]["start"] == datetime(2021, 1, 1, 10, 40)
    assert result.iloc[9]["end"] == datetime(2021, 1, 1, 10, 41)

    assert result.iloc[10]["type"] == "I"
    assert result.iloc[10]["start"] == datetime(2021, 1, 1, 10, 41)
    assert result.iloc[10]["end"] == datetime(2021, 1, 1, 10, 42)

    assert result.iloc[11]["type"] == "TN"
    assert result.iloc[11]["start"] == datetime(2021, 1, 1, 10, 42)
    assert result.iloc[11]["end"] == datetime(2021, 1, 1, 10, 45)

    assert result.iloc[12]["type"] == "Ua"
    assert result.iloc[12]["start"] == datetime(2021, 1, 1, 10, 45)
    assert result.iloc[12]["end"] == datetime(2021, 1, 1, 10, 50)

    assert result.iloc[13]["type"] == "TP"
    assert result.iloc[13]["start"] == datetime(2021, 1, 1, 10, 50)
    assert result.iloc[13]["end"] == datetime(2021, 1, 1, 10, 55)

    assert result.iloc[14]["type"] == "Oo"
    assert result.iloc[14]["start"] == datetime(2021, 1, 1, 10, 55)
    assert result.iloc[14]["end"] == datetime(2021, 1, 1, 11, 0)


def test_generate_segment_scores_with_empty_input() -> None:
    ground_truth = sf.FrameHE.from_dict({
        "lifecycle:transition": [],
        "time:timestamp": []
    })
    detected = sf.FrameHE.from_dict({
        "lifecycle:transition": [],
        "time:timestamp": []
    })
    result = ward_helper._generate_segment_scores(ground_truth, detected, None, None)
    assert len(result) == 0


def test_is_during_activity_exec() -> None:
    # Test case where timestamp is during an activity execution
    log = sf.FrameHE.from_dict({
        "lifecycle:transition": ["start", "complete"],
        "time:timestamp": [datetime(2021, 1, 1, 10, 0), datetime(2021, 1, 1, 11, 0)]
    })
    timestamp = datetime(2021, 1, 1, 10, 30)
    assert _is_during_activity_exec(log, timestamp) is True

    # Test case where timestamp is not during an activity execution
    log = sf.FrameHE.from_dict({
        "lifecycle:transition": ["start", "complete"],
        "time:timestamp": [datetime(2021, 1, 1, 10, 0), datetime(2021, 1, 1, 11, 0)]
    })
    timestamp = datetime(2021, 1, 1, 11, 30)
    assert _is_during_activity_exec(log, timestamp) is False

    # Test case where timestamp is exactly at the start of an activity
    log = sf.FrameHE.from_dict({
        "lifecycle:transition": ["start", "complete"],
        "time:timestamp": [datetime(2021, 1, 1, 10, 0), datetime(2021, 1, 1, 11, 0)]
    })
    timestamp = datetime(2021, 1, 1, 10, 0)
    assert _is_during_activity_exec(log, timestamp) is True

    # Test case where timestamp is exactly at the end of an activity
    log = sf.FrameHE.from_dict({
        "lifecycle:transition": ["start", "complete"],
        "time:timestamp": [datetime(2021, 1, 1, 10, 0), datetime(2021, 1, 1, 11, 0)]
    })
    timestamp = datetime(2021, 1, 1, 11, 0)
    assert _is_during_activity_exec(log, timestamp) is False

    # Test case where log is empty
    log = sf.FrameHE.from_dict({
        "lifecycle:transition": [],
        "time:timestamp": []
    })
    timestamp = datetime(2021, 1, 1, 10, 0)
    assert _is_during_activity_exec(log, timestamp) is False

    # Test case where lifecycle:transition is not in ["start", "complete"]
    log = sf.FrameHE.from_dict({
        "lifecycle:transition": ["other"],
        "time:timestamp": [datetime(2021, 1, 1, 10, 0)]
    })
    timestamp = datetime(2021, 1, 1, 10, 0)
    with pytest.raises(ValueError):
        _is_during_activity_exec(log, timestamp)



"""Module to calculate the 2SET metrics."""
from dataclasses import dataclass, fields
from datetime import timedelta
from functools import cached_property, lru_cache
from typing import Union, Any, Dict
import numpy as np
import static_frame as sf
from .ward_helper import _generate_segment_scores
from .utils import (_case_activity_level_metric_pre_check,
                    _generate_activity_metric_list, BUG_REPORT_CTA, _get_case_in_filtered_logs)


# pylint: disable=too-many-instance-attributes
@dataclass(frozen=True)
class TwoSet:
    """Data class to hold the 2SET metrics.

    How many of the det frames can be seen as tp, tn, d, f, ua, uo, i, m, oa, oo.
    If result of aggregated request, the values represent the average number of frames
    over the relevant case-activity pairs.
    Relative metrics are available as properties.

    :param tp: True Positives
    :param tn: True Negatives
    :param d: Deletions
    :param f: Fragmentations
    :param ua: Underfullings (at the start)
    :param uo: Underfullings (at the end)
    :param i: Insertions
    :param m: Merges
    :param oa: Overfullings (at the start)
    :param oo: Overfullings (at the end)
    """
    tp: Union[int, float]
    tn: Union[int, float]
    d: Union[int, float]
    f: Union[int, float]
    ua: Union[int, float]
    uo: Union[int, float]
    i: Union[int, float]
    m: Union[int, float]
    oa: Union[int, float]
    oo: Union[int, float]

    @cached_property
    def p(self) -> Union[int, float]:
        """Get total positive frame number in ground truth."""
        return self.tp + self.d + self.f + self.ua + self.uo

    @cached_property
    def p_det(self) -> Union[int, float]:
        """Get total positive frame number in detection."""
        return self.tp + self.i + self.m + self.oa + self.oo

    @cached_property
    def n(self) -> Union[int, float]:
        """Get total negative frames in ground truth."""
        return self.tn + self.i + self.m + self.oa + self.oo

    @cached_property
    def n_det(self) -> Union[int, float]:
        """Get total negative frames in detection."""
        return self.tn + self.d + self.f + self.ua + self.uo

    @cached_property
    def t(self) -> Union[int, float]:
        """Get total frames."""
        return (self.tp + self.tn + self.d + self.f + self.ua
                + self.uo + self.i + self.m + self.oa + self.oo)

    @cached_property
    def tpr(self) -> float:
        """Get the True Positive Rate.
        Ratio of true positives to total positives.

        Equivalent to recall. Left for compatibility."""
        return self.recall

    @cached_property
    def tnr(self) -> float:
        """Get the True Negative Rate.
        Ratio of true negatives to total negatives."""
        return round(self.tn / self.n if self.n != 0 else 0, 4)

    @cached_property
    def dr(self) -> float:
        """Get the Deletion Rate.
        Ratio of deletions to total positives."""
        return round(self.d / self.p if self.p != 0 else 0, 4)

    @cached_property
    def fr(self) -> float:
        """Get the Fragmentation Rate.
        Ratio of fragmentations to total positives."""
        return round(self.f / self.p if self.p != 0 else 0, 4)

    @cached_property
    def uar(self) -> float:
        """Get the Underfilling Rate (at the start).
        Ratio of underfullings at the start to total positives."""
        return round(self.ua / self.p if self.p != 0 else 0, 4)

    @cached_property
    def uor(self) -> float:
        """Get the Underfilling Rate (at the end).
        Ratio of underfullings at the end to total positives."""
        return round(self.uo / self.p if self.p != 0 else 0, 4)

    @cached_property
    def ir(self) -> float:
        """Get the Insertion Rate.
        Ratio of insertions to total negatives."""
        return round(self.i / self.n if self.n != 0 else 0, 4)

    @cached_property
    def mr(self) -> float:
        """Get the Merge Rate.
        Ratio of merges to total negatives."""
        return round(self.m / self.n if self.n != 0 else 0, 4)

    @cached_property
    def oar(self) -> float:
        """Get the Overfilling Rate (at the start).
        Ratio of overfullings at the start to total negatives."""
        return round(self.oa / self.n if self.n != 0 else 0, 4)

    @cached_property
    def oor(self) -> float:
        """Get the Overfilling Rate (at the end).
        Ratio of overfullings at the end to total negatives."""
        return round(self.oo / self.n if self.n != 0 else 0, 4)

    @cached_property
    def precision(self) -> float:
        """Get the precision.
        Ratio of true positives to total positive detections."""
        return round(self.tp / self.p_det if self.p_det != 0 else 0, 4)

    @cached_property
    def recall(self) -> float:
        """Get the recall.
        Ratio of true positives to total positives in ground truth."""
        return round(self.tp / self.p if self.p != 0 else 0, 4)

    @cached_property
    def f1(self) -> float:
        """Get the F1 score.
        Harmonic mean of precision and recall."""
        return round(2 * (self.precision * self.recall) / (self.precision + self.recall)
                     if self.precision + self.recall != 0 else 0, 4)

    @cached_property
    def balanced_accuracy(self) -> float:
        """Get the balanced accuracy.
        Average of true positive rate and true negative rate."""
        return round((self.tpr + self.tnr) / 2, 4)



# pylint: disable=too-many-arguments, too-many-locals, too-many-branches, too-many-statements
@lru_cache
def _two_set(gt: sf.FrameHE,
             det: sf.FrameHE,
             start_end_by_case: sf.SeriesHE[sf.Index[np.str_], Any],
             sampling_freq: float,) -> TwoSet:
    """Calculate the absolute 2SET metrics.

    Assume that logs are filtered by case and activity.
    :param gt: The ground truth DataFrame.
    :param det: The detected DataFrame.
    :param sampling_freq: The sampling frequency of the logs, in Hz.
    :param start_end_by_case: The start and end times for each case.
    :return: The absolute 2SET metrics.
    """
    case_id = _get_case_in_filtered_logs(gt, det)
    _case_activity_level_metric_pre_check(gt, det)
    sampling_timedelta = timedelta(seconds=1 / sampling_freq)
    segment_scores_frame = _generate_segment_scores(gt,
                                                    det,
                                                    start_end_by_case[case_id][0],
                                                    start_end_by_case[case_id][1])
    sorted_segments_scores_frame = segment_scores_frame.sort_values("start")
    # initialize the counters
    tp = 0
    tn = 0
    d = 0
    f = 0
    ua = 0
    uo = 0
    i = 0
    m = 0
    oa = 0
    oo = 0
    # iterate through the time represented in the segment df
    current_time = sorted_segments_scores_frame.iloc[0]["start"]
    while True:
        current_segment = sorted_segments_scores_frame.loc[
            sorted_segments_scores_frame["start"] <= current_time].iloc[-1]
        if len(current_segment) == 0:
            raise ValueError(f"No segment found for the current time. {BUG_REPORT_CTA}")
        if len(current_segment) > 1 and not isinstance(current_segment, sf.Series):
            raise ValueError("Multiple segments found for the current time. "
                             f"{BUG_REPORT_CTA}")
        # get the type of the current segment
        current_type = current_segment.loc["type"]
        # update the counters
        if current_type == "TP":
            tp += 1
        elif current_type == "TN":
            tn += 1
        elif current_type == "D":
            d += 1
        elif current_type == "F":
            f += 1
        elif current_type == "Ua":
            ua += 1
        elif current_type == "Uo":
            uo += 1
        elif current_type == "I":
            i += 1
        elif current_type == "M":
            m += 1
        elif current_type == "Oa":
            oa += 1
        elif current_type == "Oo":
            oo += 1
        # move to the next time
        current_time += sampling_timedelta
        # if the current time is larger than the last segment, break the loop
        if current_time > sorted_segments_scores_frame.iloc[-1]["end"]:
            break
    return TwoSet(tp=tp, tn=tn, d=d, f=f, ua=ua, uo=uo, i=i, m=m, oa=oa, oo=oo)


# pylint: disable=too-many-positional-arguments
def _two_set_by_activity_case(gt: sf.FrameHE,
                              det: sf.FrameHE,
                              sampling_freq: float,
                              activity_name: str,
                              case_id: str,
                              start_end_per_case: sf.SeriesHE[sf.Index[np.str_], Any]) \
                                -> TwoSet:
    """Calculate the absolute 2SET metrics for a given activity and case.

    :param gt: The ground truth log.
    :param det: The detected log.
    :param activity_name: The name of the activity.
        If "*" is passed, the 2SET metrics will be calculated
        and averaged for all activities.
    :param case_id: The case ID.
        If "*" is passed, the 2SET metrics will be calculated
        and averaged for all cases.
    :param start_end_per_case: The start and end times for each case.
    :return: The absolute 2SET metrics.
    """
    two_set_metrics = _generate_activity_metric_list(gt=gt,
                                                     det=det,
                                                     sampling_freq=sampling_freq,
                                                     case_id=case_id,
                                                     activity_name=activity_name,
                                                     start_end_per_case=start_end_per_case,
                                                     metric=_two_set)
    avg_two_set_dict: Dict[str, Union[float, int]] = {
        "tp": 0,
        "tn": 0,
        "d": 0,
        "f": 0,
        "ua": 0,
        "uo": 0,
        "i": 0,
        "m": 0,
        "oa": 0,
        "oo": 0
    }
    for two_set_metr in two_set_metrics:
        for field in fields(TwoSet):
            avg_two_set_dict[field.name] += getattr(two_set_metr, field.name)
    for field in fields(TwoSet):
        avg_two_set_dict[field.name] = round(avg_two_set_dict[field.name]
                                             / len(two_set_metrics), 4)
    return TwoSet(**avg_two_set_dict)
